import gc

import numpy as np
import pytest
import torch

from agilerl.algorithms.ddpg import DDPG
from agilerl.algorithms.dqn import DQN
from agilerl.algorithms.ppo import PPO
from agilerl.components.replay_buffer import ReplayBuffer
from agilerl.utils.probe_envs import (
    ConstantRewardContActionsDictEnv,
    ConstantRewardContActionsEnv,
    ConstantRewardContActionsImageEnv,
    ConstantRewardDictEnv,
    ConstantRewardEnv,
    ConstantRewardImageEnv,
    DiscountedRewardContActionsDictEnv,
    DiscountedRewardContActionsEnv,
    DiscountedRewardContActionsImageEnv,
    DiscountedRewardDictEnv,
    DiscountedRewardEnv,
    DiscountedRewardImageEnv,
    FixedObsPolicyContActionsDictEnv,
    FixedObsPolicyContActionsEnv,
    FixedObsPolicyContActionsImageEnv,
    FixedObsPolicyDictEnv,
    FixedObsPolicyEnv,
    FixedObsPolicyImageEnv,
    ObsDependentRewardContActionsDictEnv,
    ObsDependentRewardContActionsEnv,
    ObsDependentRewardContActionsImageEnv,
    ObsDependentRewardDictEnv,
    ObsDependentRewardEnv,
    ObsDependentRewardImageEnv,
    PolicyContActionsDictEnv,
    PolicyContActionsEnv,
    PolicyContActionsImageEnv,
    PolicyContActionsImageEnvSimple,
    PolicyDictEnv,
    PolicyEnv,
    PolicyImageEnv,
    check_policy_on_policy_with_probe_env,
    check_policy_q_learning_with_probe_env,
    check_q_learning_with_probe_env,
)


@pytest.mark.parametrize(
    "env_class, exp_state, exp_reward, exp_terminated, exp_truncated, exp_info",
    [
        (ConstantRewardEnv, 0, 1, True, False, {}),
        (ConstantRewardImageEnv, 0, 1, True, False, {}),
        (ConstantRewardContActionsEnv, 0, 1, True, False, {}),
        (ConstantRewardContActionsImageEnv, 0, 1, True, False, {}),
        (ConstantRewardDictEnv, {"discrete": 0, "box": 0}, 1, True, False, {}),
        (
            ConstantRewardContActionsDictEnv,
            {"discrete": 0, "box": 0},
            1,
            True,
            False,
            {},
        ),
    ],
)
def test_constant_reward_envs(
    env_class, exp_state, exp_reward, exp_terminated, exp_truncated, exp_info
):
    env = env_class()
    state, info = env.reset()
    for _ in range(20):
        action = [env.action_space.sample()]
        state, reward, terminated, truncated, info = env.step(action)

        if isinstance(exp_state, dict):
            assert int(np.mean(np.array(state["box"]))) == exp_state["box"]
            assert state["discrete"] == exp_state["discrete"]
        else:
            assert int(np.mean(np.array(state))) == exp_state

        assert reward == exp_reward
        assert terminated == exp_terminated
        assert truncated == exp_truncated
        assert info == exp_info

        if terminated or truncated:
            state, info = env.reset()


@pytest.mark.parametrize(
    "env_class, exp_state, exp_reward, exp_terminated, exp_truncated, exp_info",
    [
        (ObsDependentRewardEnv, 0, -1, True, False, {}),
        (ObsDependentRewardEnv, 1, 1, True, False, {}),
        (ObsDependentRewardImageEnv, 0, -1, True, False, {}),
        (ObsDependentRewardImageEnv, 1, 1, True, False, {}),
        (ObsDependentRewardDictEnv, {"discrete": 0, "box": 1}, -1, True, False, {}),
        (ObsDependentRewardDictEnv, {"discrete": 0, "box": 0}, 1, True, False, {}),
        (ObsDependentRewardDictEnv, {"discrete": 1, "box": 0}, -1, True, False, {}),
        (ObsDependentRewardDictEnv, {"discrete": 1, "box": 1}, 1, True, False, {}),
        (ObsDependentRewardContActionsEnv, 0, -1, True, False, {}),
        (ObsDependentRewardContActionsEnv, 1, 1, True, False, {}),
        (ObsDependentRewardContActionsImageEnv, 0, -1, True, False, {}),
        (ObsDependentRewardContActionsImageEnv, 1, 1, True, False, {}),
        (
            ObsDependentRewardContActionsDictEnv,
            {"discrete": 0, "box": 1},
            -1,
            True,
            False,
            {},
        ),
        (
            ObsDependentRewardContActionsDictEnv,
            {"discrete": 0, "box": 0},
            1,
            True,
            False,
            {},
        ),
        (
            ObsDependentRewardContActionsDictEnv,
            {"discrete": 1, "box": 0},
            -1,
            True,
            False,
            {},
        ),
        (
            ObsDependentRewardContActionsDictEnv,
            {"discrete": 1, "box": 1},
            1,
            True,
            False,
            {},
        ),
    ],
)
def test_observation_dependent_reward_envs(
    env_class, exp_state, exp_reward, exp_terminated, exp_truncated, exp_info
):
    env = env_class()
    state, info = env.reset()
    for _ in range(20):
        action = [env.action_space.sample()]
        state, reward, terminated, truncated, info = env.step(action)

        if isinstance(exp_state, dict):
            if all(
                [
                    int(np.mean(np.array(state[key]))) == exp_state[key]
                    for key in exp_state.keys()
                ]
            ):
                assert reward == exp_reward
        else:
            if int(np.mean(np.array(state))) == exp_state:
                assert reward == exp_reward

        assert terminated == exp_terminated
        assert truncated == exp_truncated
        assert info == exp_info

        if terminated or truncated:
            state, info = env.reset()


@pytest.mark.parametrize(
    "env_class, exp_state, exp_reward, exp_truncated, exp_info",
    [
        (DiscountedRewardEnv, 0, 0, False, {}),
        (DiscountedRewardEnv, 1, 1, False, {}),
        (DiscountedRewardImageEnv, 0, 0, False, {}),
        (DiscountedRewardImageEnv, 1, 1, False, {}),
        (DiscountedRewardDictEnv, {"discrete": 0, "box": 1}, 1, False, {}),
        (DiscountedRewardDictEnv, {"discrete": 0, "box": 0}, 0, False, {}),
        (DiscountedRewardDictEnv, {"discrete": 1, "box": 0}, 1, False, {}),
        (DiscountedRewardDictEnv, {"discrete": 1, "box": 1}, 2, False, {}),
        (DiscountedRewardContActionsEnv, 0, 0, False, {}),
        (DiscountedRewardContActionsEnv, 1, 1, False, {}),
        (DiscountedRewardContActionsImageEnv, 0, 0, False, {}),
        (DiscountedRewardContActionsImageEnv, 1, 1, False, {}),
        (DiscountedRewardContActionsDictEnv, {"discrete": 0, "box": 1}, 1, False, {}),
        (DiscountedRewardContActionsDictEnv, {"discrete": 0, "box": 0}, 0, False, {}),
        (DiscountedRewardContActionsDictEnv, {"discrete": 1, "box": 0}, 1, False, {}),
        (DiscountedRewardContActionsDictEnv, {"discrete": 1, "box": 1}, 2, False, {}),
    ],
)
def test_discounted_reward_envs(
    env_class, exp_state, exp_reward, exp_truncated, exp_info
):
    env = env_class()
    state, info = env.reset()
    for _ in range(20):
        action = [env.action_space.sample()]
        next_state, reward, terminated, truncated, info = env.step(action)

        if isinstance(exp_state, dict):
            if all(
                [
                    int(np.mean(np.array(state[key]))) == exp_state[key]
                    for key in exp_state.keys()
                ]
            ):
                assert reward == exp_reward
            assert terminated == int(np.mean(state["box"]))
        else:
            if int(np.mean(np.array(state))) == exp_state:
                assert reward == exp_reward
            assert terminated == int(np.mean(np.array(state)))
        assert truncated == exp_truncated
        assert info == exp_info

        state = next_state

        if terminated or truncated:
            state, info = env.reset()


@pytest.mark.parametrize(
    "env_class, exp_state, exp_action, exp_reward, exp_terminated, exp_truncated, exp_info",
    [
        (FixedObsPolicyEnv, 0, 0, -1, True, False, {}),
        (FixedObsPolicyEnv, 0, 1, 1, True, False, {}),
        (FixedObsPolicyImageEnv, 0, 0, -1, True, False, {}),
        (FixedObsPolicyImageEnv, 0, 1, 1, True, False, {}),
        (FixedObsPolicyDictEnv, {"discrete": 0, "box": 0}, 0, -1, True, False, {}),
        (FixedObsPolicyDictEnv, {"discrete": 0, "box": 0}, 1, 1, True, False, {}),
    ],
)
def test_discrete_actions_fixed_observation_policy_reward_envs(
    env_class,
    exp_state,
    exp_action,
    exp_reward,
    exp_terminated,
    exp_truncated,
    exp_info,
):
    env = env_class()
    state, info = env.reset()
    for _ in range(20):
        action = [env.action_space.sample()]
        state, reward, terminated, truncated, info = env.step(action)

        if isinstance(exp_state, dict):
            assert int(np.mean(np.array(state["box"]))) == exp_state["box"]
            assert state["discrete"] == exp_state["discrete"]
        else:
            assert int(np.mean(np.array(state))) == exp_state

        if action == exp_action:
            assert reward == exp_reward
        assert terminated == exp_terminated
        assert truncated == exp_truncated
        assert info == exp_info

        if terminated or truncated:
            state, info = env.reset()


@pytest.mark.parametrize(
    "env_class, exp_state, exp_action, exp_terminated, exp_truncated, exp_info",
    [
        (FixedObsPolicyContActionsEnv, 0, 1, True, False, {}),
        (FixedObsPolicyContActionsImageEnv, 0, 1, True, False, {}),
        (
            FixedObsPolicyContActionsDictEnv,
            {"discrete": 0, "box": 0},
            1,
            True,
            False,
            {},
        ),
    ],
)
def test_continuous_actions_fixed_observation_policy_reward_envs(
    env_class, exp_state, exp_action, exp_terminated, exp_truncated, exp_info
):
    env = env_class()
    state, info = env.reset()
    for _ in range(20):
        action = [env.action_space.sample()]
        state, reward, terminated, truncated, info = env.step(action)

        if isinstance(exp_state, dict):
            assert int(np.mean(np.array(state["box"]))) == exp_state["box"]
            assert state["discrete"] == exp_state["discrete"]
        else:
            assert int(np.mean(np.array(state))) == exp_state

        assert reward == -((exp_action - action[0]) ** 2)
        assert terminated == exp_terminated
        assert truncated == exp_truncated
        assert info == exp_info

        if terminated or truncated:
            state, info = env.reset()


@pytest.mark.parametrize(
    "env_class, same_reward, diff_reward, exp_terminated, exp_truncated, exp_info",
    [
        (PolicyEnv, 1, -1, True, False, {}),
        (PolicyImageEnv, 1, -1, True, False, {}),
        (PolicyDictEnv, 1, -1, True, False, {}),
    ],
)
def test_discrete_actions_policy_envs(
    env_class, same_reward, diff_reward, exp_terminated, exp_truncated, exp_info
):
    env = env_class()
    state, info = env.reset()
    for _ in range(20):
        action = [env.action_space.sample()]
        state, reward, terminated, truncated, info = env.step(action)

        if isinstance(state, dict):
            if (
                int(np.mean(np.array(state["box"]))) == action[0]
                and state["discrete"] == action[0]
            ):
                assert (
                    reward == same_reward
                ), f"action: {action}, state: {state} box: {int(np.mean(state['box']))}"
            else:
                assert reward == diff_reward
        else:
            if int(np.mean(np.array(state))) == action:
                assert reward == same_reward
            else:
                assert reward == diff_reward
        assert terminated == exp_terminated
        assert truncated == exp_truncated
        assert info == exp_info

        if terminated or truncated:
            state, info = env.reset()


@pytest.mark.parametrize(
    "env_class, reward_goal_0, reward_goal_1, exp_terminated, exp_truncated, exp_info",
    [
        (PolicyContActionsEnv, 0, 1, True, False, {}),
        (PolicyContActionsImageEnv, 0, 1, True, False, {}),
        (PolicyContActionsDictEnv, 0, 1, True, False, {}),
    ],
)
def test_continuous_actions_policy_envs(
    env_class, reward_goal_0, reward_goal_1, exp_terminated, exp_truncated, exp_info
):
    env = env_class()
    state, info = env.reset()
    for _ in range(20):
        action = env.action_space.sample()
        state, reward, terminated, truncated, info = env.step(action)

        if isinstance(state, dict):
            if int(np.mean(np.array(state["box"]))) and int(np.mean(state["discrete"])):
                assert (
                    reward
                    == -((reward_goal_0 - action[0]) ** 2)
                    - (reward_goal_1 - action[1]) ** 2
                )
            else:
                assert (
                    reward
                    == -((reward_goal_1 - action[0]) ** 2)
                    - (reward_goal_0 - action[1]) ** 2
                )
        else:
            if int(np.mean(np.array(state))):
                assert (
                    reward
                    == -((reward_goal_0 - action[0]) ** 2)
                    - (reward_goal_1 - action[1]) ** 2
                )
            else:
                assert (
                    reward
                    == -((reward_goal_1 - action[0]) ** 2)
                    - (reward_goal_0 - action[1]) ** 2
                )
        assert terminated == exp_terminated
        assert truncated == exp_truncated
        assert info == exp_info

        if terminated or truncated:
            state, info = env.reset()


@pytest.mark.parametrize(
    "env_class, reward_goal_0, reward_goal_1, exp_terminated, exp_truncated, exp_info",
    [
        (PolicyContActionsImageEnvSimple, 1, 0, True, False, {}),
    ],
)
def test_continuous_actions_policy_envs_simple(
    env_class, reward_goal_0, reward_goal_1, exp_terminated, exp_truncated, exp_info
):
    env = env_class()
    state, info = env.reset()
    for _ in range(20):
        action = env.action_space.sample()
        state, reward, terminated, truncated, info = env.step(action)

        if int(np.mean(np.array(state))):
            assert reward == -((reward_goal_0 - action[0]) ** 2)
        else:
            assert reward == -((reward_goal_1 - action[0]) ** 2)
        assert terminated == exp_terminated
        assert truncated == exp_truncated
        assert info == exp_info

        if terminated or truncated:
            state, info = env.reset()


def test_q_learning_with_probe_env():
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    env = ConstantRewardEnv()
    learn_steps = 100
    algo_args = {
        "observation_space": env.observation_space,
        "action_space": env.action_space,
        "lr": 1e-2,
    }
    memory = ReplayBuffer(
        max_size=1000,  # Max replay buffer size
        device=device,
    )
    check_q_learning_with_probe_env(env, DQN, algo_args, memory, learn_steps, device)
    gc.collect()


def test_q_learning_with_probe_env_cnn():
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    env = ConstantRewardImageEnv()
    learn_steps = 100
    algo_args = {
        "observation_space": env.observation_space,
        "action_space": env.action_space,
        "net_config": {
            "encoder_config": {
                "channel_size": [32],  # CNN channel size
                "kernel_size": [3],  # CNN kernel size
                "stride_size": [1],  # CNN stride size
            }
        },
        "normalize_images": False,
        "lr": 1e-2,
    }
    memory = ReplayBuffer(
        max_size=1000,  # Max replay buffer size
        device=device,
    )
    check_q_learning_with_probe_env(env, DQN, algo_args, memory, learn_steps, device)
    gc.collect()


def test_q_learning_with_probe_env_dict():
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    env = ConstantRewardDictEnv()
    learn_steps = 100
    algo_args = {
        "observation_space": env.observation_space,
        "action_space": env.action_space,
        "net_config": {
            "encoder_config": {
                "mlp_config": {
                    "hidden_size": [64],  # Network hidden size
                },
                "cnn_config": {
                    "channel_size": [32],  # CNN channel size
                    "kernel_size": [3],  # CNN kernel size
                    "stride_size": [1],  # CNN stride size
                },
                "latent_dim": 16,  # Latent dimension
            }
        },
        "normalize_images": False,
        "lr": 1e-2,
    }
    memory = ReplayBuffer(
        max_size=1000,  # Max replay buffer size
        device=device,
    )
    check_q_learning_with_probe_env(env, DQN, algo_args, memory, learn_steps, device)
    gc.collect()


def test_policy_q_learning_with_probe_env():
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    env = ConstantRewardContActionsEnv()
    learn_steps = 100
    algo_args = {
        "observation_space": env.observation_space,
        "action_space": env.action_space,
        "lr_actor": 1e-2,
        "lr_critic": 1e-2,
    }
    memory = ReplayBuffer(
        max_size=1000,  # Max replay buffer size
        device=device,
    )
    check_policy_q_learning_with_probe_env(
        env, DDPG, algo_args, memory, learn_steps, device
    )
    gc.collect()


def test_policy_q_learning_with_probe_env_cnn():
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    env = ConstantRewardContActionsImageEnv()
    learn_steps = 100
    algo_args = {
        "observation_space": env.observation_space,
        "action_space": env.action_space,
        "net_config": {
            "encoder_config": {
                "channel_size": [32],  # CNN channel size
                "kernel_size": [3],  # CNN kernel size
                "stride_size": [1],  # CNN stride size
            },
            "head_config": {
                "hidden_size": [64],  # Network hidden size
            },
        },
        "normalize_images": False,
        "policy_freq": 2,
        "lr_actor": 1e-2,
        "lr_critic": 1e-2,
    }
    memory = ReplayBuffer(
        max_size=1000,  # Max replay buffer size
        device=device,
    )
    check_policy_q_learning_with_probe_env(
        env, DDPG, algo_args, memory, learn_steps, device
    )
    gc.collect()


def test_policy_q_learning_with_probe_env_dict():
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    env = ConstantRewardContActionsDictEnv()
    learn_steps = 100
    algo_args = {
        "observation_space": env.observation_space,
        "action_space": env.action_space,
        "net_config": {
            "encoder_config": {
                "mlp_config": {
                    "hidden_size": [64],  # Network hidden size
                },
                "latent_dim": 16,  # Latent dimension
                "cnn_config": {
                    "channel_size": [32],  # CNN channel size
                    "kernel_size": [3],  # CNN kernel size
                    "stride_size": [1],  # CNN stride size
                },
            },
            "head_config": {
                "hidden_size": [64],  # Network hidden size
            },
        },
        "normalize_images": False,
        "policy_freq": 2,
        "lr_actor": 1e-2,
        "lr_critic": 1e-2,
    }
    memory = ReplayBuffer(
        max_size=1000,  # Max replay buffer size
        device=device,
    )
    check_policy_q_learning_with_probe_env(
        env, DDPG, algo_args, memory, learn_steps, device
    )
    gc.collect()


def test_policy_on_policy_discrete_with_probe_env():
    device = torch.device("cpu")
    env = FixedObsPolicyEnv()
    learn_steps = 10
    algo_args = {
        "observation_space": env.observation_space,
        "action_space": env.action_space,
        "lr": 0.01,
    }
    check_policy_on_policy_with_probe_env(
        env, PPO, algo_args, learn_steps, device, True
    )
    gc.collect()


def test_policy_on_policy_with_probe_env():
    device = torch.device("cpu")
    env = ConstantRewardContActionsEnv()
    learn_steps = 10
    algo_args = {
        "observation_space": env.observation_space,
        "action_space": env.action_space,
        "lr": 0.01,
    }
    check_policy_on_policy_with_probe_env(
        env, PPO, algo_args, learn_steps, device, False
    )
    gc.collect()


def test_policy_on_policy_with_probe_env_cnn():
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    env = ConstantRewardContActionsImageEnv()  # FixedObsPolicyContActionsImageEnv()
    learn_steps = 20
    algo_args = {
        "observation_space": env.observation_space,
        "action_space": env.action_space,
        "net_config": {
            "encoder_config": {
                "channel_size": [32],  # CNN channel size
                "kernel_size": [3],  # CNN kernel size
                "stride_size": [1],  # CNN stride size
            },
            "head_config": {
                "hidden_size": [64],  # Network hidden size
            },
        },
        "normalize_images": False,
        "lr": 0.01,
    }
    check_policy_on_policy_with_probe_env(
        env, PPO, algo_args, learn_steps, device, False
    )
    gc.collect()


def test_policy_on_policy_with_probe_env_dict():
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    env = ConstantRewardContActionsDictEnv()  # FixedObsPolicyContActionsDictEnv()
    learn_steps = 10
    algo_args = {
        "observation_space": env.observation_space,
        "action_space": env.action_space,
        "net_config": {
            "encoder_config": {
                "mlp_config": {
                    "hidden_size": [64],  # Network hidden size
                },
                "latent_dim": 16,  # Latent dimension
                "cnn_config": {
                    "channel_size": [32],  # CNN channel size
                    "kernel_size": [3],  # CNN kernel size
                    "stride_size": [1],  # CNN stride size
                },
            },
            "head_config": {
                "hidden_size": [64],  # Network hidden size
            },
        },
        "normalize_images": False,
        "lr": 0.01,
    }
    check_policy_on_policy_with_probe_env(
        env, PPO, algo_args, learn_steps, device, False
    )
    gc.collect()



import numpy as np
import torch

from agilerl.components.multi_agent_replay_buffer import MultiAgentReplayBuffer


# Can initialize an instance of MultiAgentReplayBuffer with valid arguments
def test_initialize_instance_with_valid_arguments():
    memory_size = 100
    field_names = ["state", "action", "reward"]
    agent_ids = ["agent1", "agent2"]

    buffer = MultiAgentReplayBuffer(memory_size, field_names, agent_ids)

    assert len(buffer) == 0
    assert buffer.memory.maxlen == memory_size
    assert buffer.field_names == field_names
    assert buffer.agent_ids == agent_ids
    assert buffer.counter == 0
    assert buffer.device is None


# Can get length of memory with __len__ method
def test_get_length_of_memory():
    memory_size = 100
    field_names = ["state", "action", "reward"]
    agent_ids = ["agent1", "agent2"]

    buffer = MultiAgentReplayBuffer(memory_size, field_names, agent_ids)

    state = {"agent1": np.array([1, 2, 3]), "agent2": np.array([1, 2, 3])}
    action = {"agent1": np.array([4, 5]), "agent2": np.array([4, 5])}
    reward = {"agent1": np.array([6]), "agent2": np.array([7])}

    # Add experiences to memory
    buffer.save_to_memory(state, action, reward)
    buffer.save_to_memory(state, action, reward)
    buffer.save_to_memory(state, action, reward)

    assert len(buffer) == 3


# Can add experiences to memory and appends to end of deque
def test_append_to_memory_deque():
    memory_size = 100
    field_names = ["state", "action", "reward"]
    agent_ids = ["agent1", "agent2"]

    buffer = MultiAgentReplayBuffer(memory_size, field_names, agent_ids)

    state = {"agent1": np.array([1, 2, 3]), "agent2": np.array([1, 2, 3])}
    action = {"agent1": np.array([4, 5]), "agent2": np.array([4, 5])}
    reward = {"agent1": np.array([6]), "agent2": np.array([7])}
    state2 = {"agent1": np.array([3, 2, 1]), "agent2": np.array([3, 2, 1])}
    action2 = {"agent1": np.array([5, 4]), "agent2": np.array([5, 4])}
    reward2 = {"agent1": np.array([7]), "agent2": np.array([6])}

    buffer._add(state, action, reward)
    buffer._add(state2, action2, reward2)

    assert len(buffer.memory) == 2
    assert buffer.memory[0].state == state
    assert buffer.memory[0].action == action
    assert buffer.memory[0].reward == reward
    assert buffer.memory[1].state == state2
    assert buffer.memory[1].action == action2
    assert buffer.memory[1].reward == reward2


# Can add an experience when memory is full and maxlen is reached
def test_add_experience_when_memory_full():
    memory_size = 2
    field_names = ["state", "action", "reward"]
    agent_ids = ["agent1", "agent2"]

    buffer = MultiAgentReplayBuffer(memory_size, field_names, agent_ids)

    state = {"agent1": np.array([1, 2, 3]), "agent2": np.array([1, 2, 3])}
    action = {"agent1": np.array([4, 5]), "agent2": np.array([4, 5])}
    reward = {"agent1": np.array([6]), "agent2": np.array([7])}
    state2 = {"agent1": np.array([3, 2, 1]), "agent2": np.array([3, 2, 1])}
    action2 = {"agent1": np.array([5, 4]), "agent2": np.array([5, 4])}
    reward2 = {"agent1": np.array([7]), "agent2": np.array([6])}
    state3 = {"agent1": np.array([7, 8, 9]), "agent2": np.array([9, 8, 7])}
    action3 = {"agent1": np.array([6, 5]), "agent2": np.array([5, 6])}
    reward3 = {"agent1": np.array([4]), "agent2": np.array([3])}

    buffer._add(state, action, reward)
    buffer._add(state2, action2, reward2)
    buffer._add(state3, action3, reward3)

    assert len(buffer.memory) == 2
    assert buffer.memory[0].state == state2
    assert buffer.memory[0].action == action2
    assert buffer.memory[0].reward == reward2
    assert buffer.memory[1].state == state3
    assert buffer.memory[1].action == action3
    assert buffer.memory[1].reward == reward3


# Can add experiences to memory using save_to_memory method
def test_add_experiences_to_memory():
    memory_size = 100
    field_names = ["state", "action", "reward"]
    agent_ids = ["agent1", "agent2"]

    buffer = MultiAgentReplayBuffer(memory_size, field_names, agent_ids)

    state = {"agent1": np.array([1, 2, 3]), "agent2": np.array([1, 2, 3])}
    action = {"agent1": np.array([4, 5]), "agent2": np.array([4, 5])}
    reward = {"agent1": np.array([6]), "agent2": np.array([7])}

    buffer.save_to_memory(state, action, reward)

    assert len(buffer) == 1
    assert buffer.memory[0].state == state
    assert buffer.memory[0].action == action
    assert buffer.memory[0].reward == reward


# Can add single experiences to memory with save_to_memory_single_env method
def test_add_single_experiences_to_memory():
    memory_size = 100
    field_names = ["state", "action", "reward"]
    agent_ids = ["agent1", "agent2"]
    device = "cpu"

    buffer = MultiAgentReplayBuffer(memory_size, field_names, agent_ids, device)

    state = {"agent1": np.array([1, 2, 3]), "agent2": np.array([1, 2, 3])}
    action = {"agent1": np.array([4, 5]), "agent2": np.array([4, 5])}
    reward = {"agent1": np.array([6]), "agent2": np.array([7])}

    buffer.save_to_memory_single_env(state, action, reward)

    assert len(buffer.memory) == 1
    assert buffer.memory[0].state == state
    assert buffer.memory[0].action == action
    assert buffer.memory[0].reward == reward


# Can add multiple experiences to memory with save_to_memory_vect_envs method
def test_add_multiple_experiences_to_memory():
    memory_size = 100
    field_names = ["state", "action", "reward"]
    agent_ids = ["agent1", "agent2"]
    device = "cpu"

    buffer = MultiAgentReplayBuffer(memory_size, field_names, agent_ids, device)

    states = {
        "agent1": np.array([[1, 2], [3, 4]]),
        "agent2": np.array([[1, 2], [3, 4]]),
    }
    actions = {"agent1": np.array([[0], [1]]), "agent2": np.array([[0], [1]])}
    rewards = {"agent1": np.array([[0], [1]]), "agent2": np.array([[0], [1]])}

    buffer.save_to_memory_vect_envs(states, actions, rewards)

    state1 = {"agent1": np.array([1, 2]), "agent2": np.array([1, 2])}
    action1 = {"agent1": np.array([0]), "agent2": np.array([0])}
    reward1 = {"agent1": np.array([0]), "agent2": np.array([0])}

    state2 = {"agent1": np.array([3, 4]), "agent2": np.array([3, 4])}
    action2 = {"agent1": np.array([1]), "agent2": np.array([1])}
    reward2 = {"agent1": np.array([1]), "agent2": np.array([1])}

    assert len(buffer.memory) == 2
    assert str(buffer.memory[0].state) == str(state1)
    assert str(buffer.memory[0].action) == str(action1)
    assert str(buffer.memory[0].reward) == str(reward1)
    assert str(buffer.memory[1].state) == str(state2)
    assert str(buffer.memory[1].action) == str(action2)
    assert str(buffer.memory[1].reward) == str(reward2)


# Can handle vectorized and un-vectorized experiences from environment
def test_add_any_experiences_to_memory():
    memory_size = 100
    field_names = ["state", "action", "reward"]
    agent_ids = ["agent1", "agent2"]
    device = "cpu"

    buffer = MultiAgentReplayBuffer(memory_size, field_names, agent_ids, device)

    states = {
        "agent1": np.array([[1, 2], [3, 4]]),
        "agent2": np.array([[1, 2], [3, 4]]),
    }
    actions = {"agent1": np.array([[0], [1]]), "agent2": np.array([[0], [1]])}
    rewards = {"agent1": np.array([[0], [1]]), "agent2": np.array([[0], [1]])}

    buffer.save_to_memory(states, actions, rewards, is_vectorised=True)

    state1 = {"agent1": np.array([1, 2]), "agent2": np.array([1, 2])}
    action1 = {"agent1": np.array([0]), "agent2": np.array([0])}
    reward1 = {"agent1": np.array([0]), "agent2": np.array([0])}

    state2 = {"agent1": np.array([3, 4]), "agent2": np.array([3, 4])}
    action2 = {"agent1": np.array([1]), "agent2": np.array([1])}
    reward2 = {"agent1": np.array([1]), "agent2": np.array([1])}

    assert len(buffer.memory) == 2
    assert str(buffer.memory[0].state) == str(state1)
    assert str(buffer.memory[0].action) == str(action1)
    assert str(buffer.memory[0].reward) == str(reward1)
    assert str(buffer.memory[1].state) == str(state2)
    assert str(buffer.memory[1].action) == str(action2)
    assert str(buffer.memory[1].reward) == str(reward2)

    new_state = {"agent1": np.array([1, 2]), "agent2": np.array([1, 2])}
    new_action = {"agent1": np.array([0]), "agent2": np.array([0])}
    new_reward = {"agent1": np.array([0]), "agent2": np.array([0])}

    buffer.save_to_memory(new_state, new_action, new_reward, is_vectorised=False)

    assert len(buffer.memory) == 3
    assert str(buffer.memory[2].state) == str(new_state)
    assert str(buffer.memory[2].action) == str(new_action)
    assert str(buffer.memory[2].reward) == str(new_reward)


# Test that the sampled experiences align with the correct environment
def test_add_and_sample_vect_experiences_to_memory():
    memory_size = 100
    field_names = ["environments", "state"]
    agent_ids = ["agent1", "agent2"]
    device = "cpu"
    buffer = MultiAgentReplayBuffer(memory_size, field_names, agent_ids, device)
    num_envs = 50
    environments = {
        agent: np.array([[env] for env in range(num_envs)]) for agent in agent_ids
    }
    states = {
        agent: np.array([[(env + 1) * (a_idx + 1)] for env in range(num_envs)])
        for a_idx, agent in enumerate(agent_ids)
    }
    buffer.save_to_memory(states, environments, is_vectorised=True)
    sampled_states, sampled_environments = buffer.sample(batch_size=25)
    for agent in sampled_environments:
        for idx, env in enumerate(sampled_environments[agent]):
            assert states[agent][int(env)] == sampled_states[agent][idx].cpu().numpy()


# Can sample experiences from memory using sample method
def test_sample_experiences_from_memory():
    memory_size = 100
    field_names = ["state", "action", "reward"]
    agent_ids = ["agent1", "agent2"]

    buffer = MultiAgentReplayBuffer(memory_size, field_names, agent_ids)

    state = {"agent1": np.array([1, 2, 3]), "agent2": np.array([1, 2, 3])}
    action = {"agent1": np.array([4, 5]), "agent2": np.array([4, 5])}
    reward = {"agent1": np.array([6]), "agent2": np.array([7])}

    buffer.save_to_memory(state, action, reward)

    batch_size = 1
    transition = buffer.sample(batch_size)

    assert len(transition) == len(field_names)
    assert transition[0]["agent1"].shape == (batch_size,) + state["agent1"].shape
    assert transition[1]["agent1"].shape == (batch_size,) + action["agent1"].shape
    assert transition[2]["agent1"].shape == (batch_size,) + reward["agent1"].shape


# Can sample experiences from memory using sample method
def test_sample_experiences_from_memory_images():
    memory_size = 100
    field_names = ["state", "action", "reward"]
    agent_ids = ["agent1", "agent2"]

    buffer = MultiAgentReplayBuffer(memory_size, field_names, agent_ids)

    state = {
        "agent1": np.random.rand(3, 128, 128),
        "agent2": np.random.rand(3, 128, 128),
    }
    action = {"agent1": np.array([4, 5]), "agent2": np.array([4, 5])}
    reward = {"agent1": np.array([6]), "agent2": np.array([7])}

    buffer.save_to_memory(state, action, reward)

    batch_size = 1
    transition = buffer.sample(batch_size)

    assert len(transition) == len(field_names)
    assert transition[0]["agent1"].shape == (batch_size,) + state["agent1"].shape
    assert transition[1]["agent1"].shape == (batch_size,) + action["agent1"].shape
    assert transition[2]["agent1"].shape == (batch_size,) + reward["agent1"].shape


# Can sample experiences from memory using sample method
def test_sample_experiences_from_memory_dict():
    memory_size = 100
    field_names = ["state", "action", "reward"]
    agent_ids = ["agent1", "agent2"]

    buffer = MultiAgentReplayBuffer(memory_size, field_names, agent_ids)

    state = {
        "agent1": {i: np.random.rand(3, 128, 128) for i in range(3)},
        "agent2": {i: np.random.rand(3, 128, 128) for i in range(3)},
    }
    action = {"agent1": np.array([4, 5]), "agent2": np.array([4, 5])}
    reward = {"agent1": np.array([6]), "agent2": np.array([7])}

    buffer.save_to_memory(state, action, reward)

    batch_size = 1
    transition = buffer.sample(batch_size)

    assert len(transition) == len(field_names)
    assert transition[1]["agent1"].shape == (batch_size,) + action["agent1"].shape
    assert transition[2]["agent1"].shape == (batch_size,) + reward["agent1"].shape


# Can process a transition from experiences and return a dictionary of numpy arrays.
def test_returns_np_transition_dictionary():
    memory_size = 100
    field_names = ["state", "action", "reward", "next_state", "done"]
    agent_ids = ["agent1", "agent2"]

    # Initialize the class object
    buffer = MultiAgentReplayBuffer(
        memory_size=memory_size, field_names=field_names, agent_ids=agent_ids
    )

    # Create some dummy experiences
    experience1 = buffer.experience(
        state={"agent1": np.array([1]), "agent2": np.array([4])},
        action={"agent1": np.array([0]), "agent2": np.array([1])},
        reward={"agent1": np.array([0]), "agent2": np.array([1])},
        next_state={"agent1": np.array([3]), "agent2": np.array([6])},
        done={"agent1": np.array([True]), "agent2": np.array([False])},
    )
    experience2 = buffer.experience(
        state={"agent1": np.array([4]), "agent2": np.array([7])},
        action={"agent1": np.array([1]), "agent2": np.array([0])},
        reward={"agent1": np.array([1]), "agent2": np.array([0])},
        next_state={"agent1": np.array([6]), "agent2": np.array([9])},
        done={"agent1": np.array([False]), "agent2": np.array([True])},
    )
    experiences = [experience1, experience2]

    # Call the method under test
    transition = buffer._process_transition(experiences, np_array=True)

    # Check the transition dictionary returned
    assert isinstance(transition, dict)
    assert set(transition.keys()) == set(field_names)
    assert isinstance(transition["state"], dict)
    assert isinstance(transition["action"], dict)
    assert isinstance(transition["reward"], dict)
    assert isinstance(transition["next_state"], dict)
    assert isinstance(transition["done"], dict)
    assert set(transition["state"].keys()) == set(agent_ids)
    assert set(transition["action"].keys()) == set(agent_ids)
    assert set(transition["reward"].keys()) == set(agent_ids)
    assert set(transition["next_state"].keys()) == set(agent_ids)
    assert set(transition["done"].keys()) == set(agent_ids)
    assert isinstance(transition["state"]["agent1"], np.ndarray)
    assert isinstance(transition["state"]["agent2"], np.ndarray)
    assert isinstance(transition["action"]["agent1"], np.ndarray)
    assert isinstance(transition["action"]["agent2"], np.ndarray)
    assert isinstance(transition["reward"]["agent1"], np.ndarray)
    assert isinstance(transition["reward"]["agent2"], np.ndarray)
    assert isinstance(transition["next_state"]["agent1"], np.ndarray)
    assert isinstance(transition["next_state"]["agent2"], np.ndarray)
    assert isinstance(transition["done"]["agent1"], np.ndarray)
    assert isinstance(transition["done"]["agent2"], np.ndarray)
    assert transition["state"]["agent1"].shape == (len(experiences), 1)
    assert transition["action"]["agent1"].shape == (len(experiences), 1)
    assert transition["reward"]["agent1"].shape == (len(experiences), 1)
    assert transition["next_state"]["agent1"].shape == (len(experiences), 1)
    assert transition["done"]["agent1"].shape == (len(experiences), 1)
    assert np.array_equal(transition["state"]["agent1"], np.array([[1], [4]]))
    assert np.array_equal(transition["state"]["agent2"], np.array([[4], [7]]))
    assert np.array_equal(transition["action"]["agent1"], np.array([[0], [1]]))
    assert np.array_equal(transition["action"]["agent2"], np.array([[1], [0]]))
    assert np.array_equal(transition["reward"]["agent1"], np.array([[0], [1]]))
    assert np.array_equal(transition["reward"]["agent2"], np.array([[1], [0]]))
    assert np.array_equal(transition["next_state"]["agent1"], np.array([[3], [6]]))
    assert np.array_equal(transition["next_state"]["agent2"], np.array([[6], [9]]))
    assert np.array_equal(transition["done"]["agent1"], np.array([[1], [0]]))
    assert np.array_equal(transition["done"]["agent2"], np.array([[0], [1]]))


# Can process a transition from a single experience and return a dictionary of numpy arrays.
def test_returns_np_transition_asingle_experience_dictionary():
    memory_size = 100
    field_names = ["state", "action", "reward", "next_state", "done"]
    agent_ids = ["agent1", "agent2"]

    # Initialize the class object
    buffer = MultiAgentReplayBuffer(
        memory_size=memory_size, field_names=field_names, agent_ids=agent_ids
    )

    # Create some dummy experiences
    experience1 = buffer.experience(
        state={"agent1": np.array([1]), "agent2": np.array([4])},
        action={"agent1": np.array([0]), "agent2": np.array([1])},
        reward={"agent1": np.array([0]), "agent2": np.array([1])},
        next_state={"agent1": np.array([3]), "agent2": np.array([6])},
        done={"agent1": np.array([True]), "agent2": np.array([False])},
    )
    experiences = [experience1]

    # Call the method under test
    transition = buffer._process_transition(experiences, np_array=True)

    # Check the transition dictionary returned
    assert isinstance(transition, dict)
    assert set(transition.keys()) == set(field_names)
    assert isinstance(transition["state"], dict)
    assert isinstance(transition["action"], dict)
    assert isinstance(transition["reward"], dict)
    assert isinstance(transition["next_state"], dict)
    assert isinstance(transition["done"], dict)
    assert set(transition["state"].keys()) == set(agent_ids)
    assert set(transition["action"].keys()) == set(agent_ids)
    assert set(transition["reward"].keys()) == set(agent_ids)
    assert set(transition["next_state"].keys()) == set(agent_ids)
    assert set(transition["done"].keys()) == set(agent_ids)
    assert isinstance(transition["state"]["agent1"], np.ndarray)
    assert isinstance(transition["state"]["agent2"], np.ndarray)
    assert isinstance(transition["action"]["agent1"], np.ndarray)
    assert isinstance(transition["action"]["agent2"], np.ndarray)
    assert isinstance(transition["reward"]["agent1"], np.ndarray)
    assert isinstance(transition["reward"]["agent2"], np.ndarray)
    assert isinstance(transition["next_state"]["agent1"], np.ndarray)
    assert isinstance(transition["next_state"]["agent2"], np.ndarray)
    assert isinstance(transition["done"]["agent1"], np.ndarray)
    assert isinstance(transition["done"]["agent2"], np.ndarray)
    assert transition["state"]["agent1"].shape == (len(experiences), 1)
    assert transition["action"]["agent1"].shape == (len(experiences), 1)
    assert transition["reward"]["agent1"].shape == (len(experiences), 1)
    assert transition["next_state"]["agent1"].shape == (len(experiences), 1)
    assert transition["done"]["agent1"].shape == (len(experiences), 1)
    assert np.array_equal(transition["state"]["agent1"], np.array([[1]]))
    assert np.array_equal(transition["state"]["agent2"], np.array([[4]]))
    assert np.array_equal(transition["action"]["agent1"], np.array([[0]]))
    assert np.array_equal(transition["action"]["agent2"], np.array([[1]]))
    assert np.array_equal(transition["reward"]["agent1"], np.array([[0]]))
    assert np.array_equal(transition["reward"]["agent2"], np.array([[1]]))
    assert np.array_equal(transition["next_state"]["agent1"], np.array([[3]]))
    assert np.array_equal(transition["next_state"]["agent2"], np.array([[6]]))
    assert np.array_equal(transition["done"]["agent1"], np.array([[1]]))
    assert np.array_equal(transition["done"]["agent2"], np.array([[0]]))


# Can process a transition from experiences and return a dictionary of numpy arrays.
def test_returns_np_transition_dictionary_images():
    memory_size = 100
    field_names = ["state", "action", "reward", "next_state", "done"]
    agent_ids = ["agent1", "agent2"]

    # Initialize the class object
    buffer = MultiAgentReplayBuffer(
        memory_size=memory_size, field_names=field_names, agent_ids=agent_ids
    )

    # Create some dummy experiences
    experience1 = buffer.experience(
        state={"agent1": np.random.rand(3, 128, 128), "agent2": np.array([4])},
        action={"agent1": np.array([0]), "agent2": np.array([1])},
        reward={"agent1": np.array([0]), "agent2": np.array([1])},
        next_state={"agent1": np.random.rand(3, 128, 128), "agent2": np.array([6])},
        done={"agent1": np.array([True]), "agent2": np.array([False])},
    )
    experience2 = buffer.experience(
        state={"agent1": np.random.rand(3, 128, 128), "agent2": np.array([7])},
        action={"agent1": np.array([1]), "agent2": np.array([0])},
        reward={"agent1": np.array([1]), "agent2": np.array([0])},
        next_state={"agent1": np.random.rand(3, 128, 128), "agent2": np.array([9])},
        done={"agent1": np.array([False]), "agent2": np.array([True])},
    )
    experiences = [experience1, experience2]

    # Call the method under test
    transition = buffer._process_transition(experiences, np_array=True)

    # Check the transition dictionary returned
    assert isinstance(transition, dict)
    assert set(transition.keys()) == set(field_names)
    assert isinstance(transition["state"], dict)
    assert isinstance(transition["action"], dict)
    assert isinstance(transition["reward"], dict)
    assert isinstance(transition["next_state"], dict)
    assert isinstance(transition["done"], dict)
    assert set(transition["state"].keys()) == set(agent_ids)
    assert set(transition["action"].keys()) == set(agent_ids)
    assert set(transition["reward"].keys()) == set(agent_ids)
    assert set(transition["next_state"].keys()) == set(agent_ids)
    assert set(transition["done"].keys()) == set(agent_ids)
    assert isinstance(transition["state"]["agent1"], np.ndarray)
    assert isinstance(transition["state"]["agent2"], np.ndarray)
    assert isinstance(transition["action"]["agent1"], np.ndarray)
    assert isinstance(transition["action"]["agent2"], np.ndarray)
    assert isinstance(transition["reward"]["agent1"], np.ndarray)
    assert isinstance(transition["reward"]["agent2"], np.ndarray)
    assert isinstance(transition["next_state"]["agent1"], np.ndarray)
    assert isinstance(transition["next_state"]["agent2"], np.ndarray)
    assert isinstance(transition["done"]["agent1"], np.ndarray)
    assert isinstance(transition["done"]["agent2"], np.ndarray)
    assert transition["state"]["agent1"].shape == (len(experiences), 3, 128, 128)
    assert transition["action"]["agent1"].shape == (len(experiences), 1)
    assert transition["reward"]["agent1"].shape == (len(experiences), 1)
    assert transition["next_state"]["agent1"].shape == (len(experiences), 3, 128, 128)
    assert transition["done"]["agent1"].shape == (len(experiences), 1)
    assert np.array_equal(transition["action"]["agent1"], np.array([[0], [1]]))
    assert np.array_equal(transition["action"]["agent2"], np.array([[1], [0]]))
    assert np.array_equal(transition["reward"]["agent1"], np.array([[0], [1]]))
    assert np.array_equal(transition["reward"]["agent2"], np.array([[1], [0]]))
    assert np.array_equal(transition["done"]["agent1"], np.array([[1], [0]]))
    assert np.array_equal(transition["done"]["agent2"], np.array([[0], [1]]))


# Can process a transition from experiences and return a dictionary of torch tensors.
def test_returns_torch_transition_dictionary():
    memory_size = 100
    field_names = ["state", "action", "reward", "next_state", "done"]
    agent_ids = ["agent1", "agent2"]
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    # Initialize the class object
    buffer = MultiAgentReplayBuffer(
        memory_size=memory_size,
        field_names=field_names,
        agent_ids=agent_ids,
        device=device,
    )

    # Create some dummy experiences
    experience1 = buffer.experience(
        state={"agent1": np.array([1, 2, 3]), "agent2": np.array([4, 5, 6])},
        action={"agent1": np.array([0]), "agent2": np.array([1])},
        reward={"agent1": np.array([0]), "agent2": np.array([1])},
        next_state={"agent1": np.array([3, 2, 1]), "agent2": np.array([6, 5, 4])},
        done={"agent1": np.array([True]), "agent2": np.array([False])},
    )
    experience2 = buffer.experience(
        state={"agent1": np.array([4, 5, 6]), "agent2": np.array([7, 8, 9])},
        action={"agent1": np.array([1]), "agent2": np.array([0])},
        reward={"agent1": np.array([1]), "agent2": np.array([0])},
        next_state={"agent1": np.array([6, 5, 4]), "agent2": np.array([9, 8, 7])},
        done={"agent1": np.array([False]), "agent2": np.array([True])},
    )
    experiences = [experience1, experience2]

    # Call the method under test
    transition = buffer._process_transition(experiences)

    # Check the transition dictionary returned
    assert isinstance(transition, dict)
    assert set(transition.keys()) == set(field_names)
    assert isinstance(transition["state"], dict)
    assert isinstance(transition["action"], dict)
    assert isinstance(transition["reward"], dict)
    assert isinstance(transition["next_state"], dict)
    assert isinstance(transition["done"], dict)
    assert set(transition["state"].keys()) == set(agent_ids)
    assert set(transition["action"].keys()) == set(agent_ids)
    assert set(transition["reward"].keys()) == set(agent_ids)
    assert set(transition["next_state"].keys()) == set(agent_ids)
    assert set(transition["done"].keys()) == set(agent_ids)
    assert isinstance(transition["state"]["agent1"], torch.Tensor)
    assert isinstance(transition["state"]["agent2"], torch.Tensor)
    assert isinstance(transition["action"]["agent1"], torch.Tensor)
    assert isinstance(transition["action"]["agent2"], torch.Tensor)
    assert isinstance(transition["reward"]["agent1"], torch.Tensor)
    assert isinstance(transition["reward"]["agent2"], torch.Tensor)
    assert isinstance(transition["next_state"]["agent1"], torch.Tensor)
    assert isinstance(transition["next_state"]["agent2"], torch.Tensor)
    assert isinstance(transition["done"]["agent1"], torch.Tensor)
    assert isinstance(transition["done"]["agent2"], torch.Tensor)
    assert transition["state"]["agent1"].shape == (len(experiences), 3)
    assert transition["action"]["agent1"].shape == (len(experiences), 1)
    assert transition["reward"]["agent1"].shape == (len(experiences), 1)
    assert transition["next_state"]["agent1"].shape == (len(experiences), 3)
    assert transition["done"]["agent1"].shape == (len(experiences), 1)
    assert torch.equal(
        transition["state"]["agent1"],
        torch.from_numpy(np.array([[1, 2, 3], [4, 5, 6]])).to(device),
    )
    assert torch.equal(
        transition["state"]["agent2"],
        torch.from_numpy(np.array([[4, 5, 6], [7, 8, 9]])).to(device),
    )
    assert torch.equal(
        transition["action"]["agent1"],
        torch.from_numpy(np.array([[0], [1]])).to(device),
    )
    assert torch.equal(
        transition["action"]["agent2"],
        torch.from_numpy(np.array([[1], [0]])).to(device),
    )
    assert torch.equal(
        transition["reward"]["agent1"],
        torch.from_numpy(np.array([[0], [1]])).to(device),
    )
    assert torch.equal(
        transition["reward"]["agent2"],
        torch.from_numpy(np.array([[1], [0]])).to(device),
    )
    assert torch.equal(
        transition["next_state"]["agent1"],
        torch.from_numpy(np.array([[3, 2, 1], [6, 5, 4]])).to(device),
    )
    assert torch.equal(
        transition["next_state"]["agent2"],
        torch.from_numpy(np.array([[6, 5, 4], [9, 8, 7]])).to(device),
    )
    assert torch.equal(
        transition["done"]["agent1"], torch.from_numpy(np.array([[1], [0]])).to(device)
    )
    assert torch.equal(
        transition["done"]["agent2"], torch.from_numpy(np.array([[0], [1]])).to(device)
    )



import torch

from agilerl.modules.custom_components import GumbelSoftmax


def test_apply_gumbel_softmax():
    # Initialize the GumbelSoftmax class
    gumbel_softmax = GumbelSoftmax()

    # Create input tensor
    logits = torch.tensor([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])

    # Apply gumbel softmax function
    output = gumbel_softmax(logits)

    # Check if output tensor has the same shape as input tensor
    assert output.shape == logits.shape

    # Check if output tensor is within the range [0, 1]
    assert torch.all(output >= 0) and torch.all(output <= 1)

    # Check if output tensor sums up to 1 along the last dimension
    assert torch.all(torch.isclose(torch.sum(output, dim=-1), torch.tensor([1.0, 1.0])))



import warnings
from collections import OrderedDict
from typing import Any, Generator, Optional, Union

import numpy as np
import torch
from gymnasium import spaces
from tensordict import TensorDict, TensorDictBase, is_tensor_collection

from agilerl.typing import (
    ArrayOrTensor,
    BPTTSequenceType,
    ObservationType,
    StandardTensorDict,
    TorchObsType,
)
from agilerl.utils.algo_utils import (
    extract_sequences_from_episode,
    get_num_actions,
    get_obs_shape,
    maybe_add_batch_dim,
)


class RolloutBuffer:
    """
    Rollout buffer for collecting experiences and computing advantages for RL algorithms.
    This buffer is designed to handle vectorized environments efficiently.

    :param capacity: Maximum number of timesteps to store in the buffer (per environment).
    :type capacity: int
    :param observation_space: Observation space of the environment.
    :type observation_space: gym.spaces.Space
    :param action_space: Action space of the environment.
    :type action_space: gym.spaces.Space
    :param num_envs: Number of parallel environments.
    :type num_envs: int
    :param device: Device to store tensors on, defaults to "cpu".
    :type device: str, optional
    :param gae_lambda: Lambda parameter for GAE, defaults to 0.95.
    :type gae_lambda: float, optional
    :param gamma: Discount factor, defaults to 0.99.
    :type gamma: float, optional
    :param recurrent: Whether to store hidden states, defaults to False.
    :type recurrent: bool, optional
    :param hidden_state_architecture: Architecture of hidden states if used, defaults to None.
    :type hidden_state_architecture: dict[str, tuple[int, int, int]], optional
    :param use_gae: Whether to compute GAE advantages, defaults to True.
    :type use_gae: bool, optional
    :param wrap_at_capacity: Whether to wrap the buffer at capacity, defaults to False. This is especially useful
        for OFF-policy algorithms, ON-policy algorithms should leave this as False in most cases.
    :type wrap_at_capacity: bool, optional
    :param max_seq_len: Maximum sequence length for BPTT, defaults to None.
    :type max_seq_len: int, optional
    """

    padded_data: TensorDict
    unpadded_data: TensorDict
    unpadded_slices: list[torch.Tensor]

    def __init__(
        self,
        capacity: int,
        observation_space: spaces.Space,
        action_space: spaces.Space,
        num_envs: int = 1,
        device: str = "cpu",
        gae_lambda: float = 0.95,
        gamma: float = 0.99,
        recurrent: bool = False,
        hidden_state_architecture: Optional[dict[str, tuple[int, int, int]]] = None,
        use_gae: bool = True,
        wrap_at_capacity: bool = False,
        max_seq_len: Optional[int] = None,
        bptt_sequence_type: BPTTSequenceType = BPTTSequenceType.CHUNKED,
    ):
        self.capacity = capacity
        self.observation_space = observation_space
        self.action_space = action_space
        self.num_envs = num_envs
        self.device = device
        self.gamma = gamma
        self.gae_lambda = gae_lambda
        self.recurrent = recurrent
        self.hidden_state_architecture = hidden_state_architecture
        self.use_gae = use_gae
        self.wrap_at_capacity = wrap_at_capacity
        self.max_seq_len = max_seq_len
        self.bptt_sequence_type = bptt_sequence_type

        self.pos = 0
        self.full = False
        self.num_sequences = None
        self.max_sequence_length = None
        self.unpadded_slices = None
        self.padded_data = None
        self.unpadded_data = None
        self._initialize_buffers()

    def size(self) -> int:
        """
        Get current number of transitions stored in the buffer.

        :return: Current number of transitions.
        :rtype: int
        """
        return (self.capacity if self.full else self.pos) * self.num_envs

    def reset(self) -> None:
        """Reset the buffer pointer and full flag."""
        self.pos = 0
        self.full = False
        self.num_sequences = None
        self.max_sequence_length = None
        self.unpadded_slices = None
        self.padded_data = None
        self.unpadded_data = None

    def _maybe_reshape_obs(
        self, obs: TorchObsType, space: spaces.Space
    ) -> TorchObsType:
        """Reshape observation to the correct shape.

        :param obs: Observation to reshape.
        :type obs: TorchObsType
        :param space: Observation space.
        :type space: spaces.Space
        :return: Reshaped observation.
        :rtype: TorchObsType
        """
        if isinstance(space, spaces.Discrete) and obs.ndim < 2:
            obs = obs.unsqueeze(-1)

        return maybe_add_batch_dim(obs, space)

    def _initialize_buffers(self) -> None:
        """Initialize buffer arrays with correct shapes for vectorized environments."""
        # Determine shapes and dtypes for all expected fields
        obs_shape = get_obs_shape(self.observation_space)
        num_actions = get_num_actions(self.action_space)

        # Create a source TensorDict with appropriately sized tensors
        # The tensors will be on the CPU by default, can be moved to device later if needed.
        source_dict = OrderedDict()
        if isinstance(
            self.observation_space, spaces.Dict
        ):  # Nested structure for Dict spaces
            obs_dict = OrderedDict()
            for key, shape in obs_shape.items():
                obs_dict[key] = torch.zeros(
                    (self.capacity, self.num_envs, *shape), dtype=torch.float32
                )

            source_dict["observations"] = obs_dict
            source_dict["next_observations"] = {
                key: torch.zeros_like(tensor) for key, tensor in obs_dict.items()
            }
        else:
            # For non-Dict spaces, use regular tensor allocation
            source_dict["observations"] = torch.zeros(
                (self.capacity, self.num_envs, *obs_shape), dtype=torch.float32
            )
            source_dict["next_observations"] = torch.zeros(
                (self.capacity, self.num_envs, *obs_shape), dtype=torch.float32
            )

        # Add other standard fields
        source_dict.update(
            {
                "actions": torch.zeros(
                    (self.capacity, self.num_envs, num_actions), dtype=torch.float32
                ),
                "rewards": torch.zeros(
                    (self.capacity, self.num_envs), dtype=torch.float32
                ),
                "dones": torch.zeros((self.capacity, self.num_envs), dtype=torch.bool),
                "values": torch.zeros(
                    (self.capacity, self.num_envs), dtype=torch.float32
                ),
                "log_probs": torch.zeros(
                    (self.capacity, self.num_envs), dtype=torch.float32
                ),
                "advantages": torch.zeros(
                    (self.capacity, self.num_envs), dtype=torch.float32
                ),
                "returns": torch.zeros(
                    (self.capacity, self.num_envs), dtype=torch.float32
                ),
            }
        )

        if self.recurrent:
            if self.hidden_state_architecture is None:
                raise ValueError(
                    "hidden_state_architecture must be provided if recurrent=True"
                )
            # self.hidden_state_architecture is dict[str, tuple[num_layers, num_envs_at_ppo_init, hidden_size]]
            # For buffer storage, each hidden state key should map to a tensor of shape:
            # (capacity, num_envs, num_layers, hidden_size)
            source_dict["hidden_states"] = {
                key: torch.zeros(
                    (
                        self.capacity,
                        self.num_envs,
                        self.hidden_state_architecture[key][0],  # num_layers
                        self.hidden_state_architecture[key][2],  # hidden_size
                    ),
                    dtype=torch.float32,
                )
                for key in self.hidden_state_architecture.keys()
            }
            # `next_hidden_states` might not be strictly necessary if we only store initial hidden states
            # for sequences. If we store step-by-step next_hidden_states, it would be:
            # source_dict["next_hidden_states"] = { ... similar structure ... }

        # Initialize the main buffer as a TensorDict with batch_size [capacity, num_envs]
        self.buffer = TensorDict(
            source_dict,
            batch_size=[self.capacity, self.num_envs],
            device="cpu",  # Keep buffer on CPU for memory efficiency, move to device in get_tensor_batch
        )

    def add(
        self,
        obs: ObservationType,
        action: ArrayOrTensor,
        reward: Union[float, np.ndarray],
        done: Union[bool, np.ndarray],
        value: Union[float, np.ndarray],
        log_prob: Union[float, np.ndarray],
        next_obs: Optional[ObservationType] = None,
        hidden_state: Optional[dict[str, ArrayOrTensor]] = None,
        next_hidden_state: Optional[
            dict[str, ArrayOrTensor]
        ] = None,  # Not used if only initial hidden states are stored
        episode_start: Optional[Union[bool, np.ndarray]] = None,
    ) -> None:
        """
        Add a new batch of observations and associated data from vectorized environments to the buffer.

        :param obs: Current observation batch (shape: (num_envs, *obs_shape))
        :type obs: ObservationType
        :param action: Action batch taken (shape: (num_envs, *action_shape))
        :type action: ArrayOrTensor
        :param reward: Reward batch received (shape: (num_envs,))
        :type reward: Union[float, np.ndarray]
        :param done: Done flag batch (shape: (num_envs,))
        :type done: Union[bool, np.ndarray]
        :param value: Value estimate batch (shape: (num_envs,))
        :type value: Union[float, np.ndarray]
        :param log_prob: Log probability batch of the actions (shape: (num_envs,))
        :type log_prob: Union[float, np.ndarray]
        :param next_obs: Next observation batch (shape: (num_envs, *obs_shape)), defaults to None
        :type next_obs: Optional[ObservationType]
        :param hidden_state: Current hidden state batch (shape: (num_envs, hidden_size)), defaults to None
        :type hidden_state: Optional[dict[str, ArrayOrTensor]]
        :param next_hidden_state: Next hidden state batch (shape: (num_envs, hidden_size)), defaults to None
        :type next_hidden_state: Optional[dict[str, ArrayOrTensor]]
        :param episode_start: Episode start flag batch (shape: (num_envs,)), defaults to None
        :type episode_start: Optional[Union[bool, np.ndarray]]
        """
        if self.pos == self.capacity:
            if self.wrap_at_capacity:
                self.pos = 0
            else:
                raise ValueError(
                    f"Buffer has reached capacity ({self.capacity} transitions) but received more transitions. "
                    "Either increase buffer capacity or set wrap_at_capacity=True."
                )

        # Prepare data as a dictionary of tensors for the current time step
        current_step_data = OrderedDict()

        # Convert inputs to tensors and ensure correct device (CPU for buffer storage)
        # Also ensure they have the (num_envs, ...) shape
        if isinstance(self.observation_space, spaces.Dict):
            obs_dict = OrderedDict()
            for key, item in obs.items():
                sub_space = self.observation_space.spaces[key]
                obs_tensor = torch.as_tensor(item, device="cpu")
                obs_dict[key] = self._maybe_reshape_obs(obs_tensor, sub_space)

            current_step_data["observations"] = obs_dict
        else:
            obs_tensor = torch.as_tensor(obs, device="cpu")
            current_step_data["observations"] = self._maybe_reshape_obs(
                obs_tensor, self.observation_space
            )

        # Actions
        action_tensor = torch.as_tensor(action, device="cpu")
        current_step_data["actions"] = action_tensor.reshape(self.num_envs, -1)

        # Rewards
        reward_tensor = torch.as_tensor(reward, dtype=torch.float32, device="cpu")
        current_step_data["rewards"] = reward_tensor.reshape(self.num_envs)

        # Dones
        done_tensor = torch.as_tensor(done, dtype=torch.bool, device="cpu")
        current_step_data["dones"] = done_tensor.reshape(self.num_envs)

        # Values
        value_tensor = torch.as_tensor(value, dtype=torch.float32, device="cpu")
        current_step_data["values"] = value_tensor.reshape(self.num_envs)

        # Log_probs
        log_prob_tensor = torch.as_tensor(log_prob, dtype=torch.float32, device="cpu")
        current_step_data["log_probs"] = log_prob_tensor.reshape(self.num_envs)

        # Next Observations
        if next_obs is not None:
            if isinstance(self.observation_space, spaces.Dict):
                next_obs_dict = OrderedDict()
                for key, item in next_obs.items():
                    sub_space = self.observation_space.spaces[key]
                    next_obs_tensor = torch.as_tensor(item, device="cpu")
                    next_obs_dict[key] = self._maybe_reshape_obs(
                        next_obs_tensor, sub_space
                    )

                current_step_data["next_observations"] = next_obs_dict
            else:
                next_obs_tensor = torch.as_tensor(next_obs, device="cpu")
                current_step_data["next_observations"] = self._maybe_reshape_obs(
                    next_obs_tensor, self.observation_space
                )

        # Episode Starts
        if episode_start is not None:
            episode_start_tensor = torch.as_tensor(
                episode_start, dtype=torch.bool, device="cpu"
            )
            current_step_data["episode_starts"] = episode_start_tensor.reshape(
                self.num_envs
            )
        else:
            current_step_data["episode_starts"] = torch.zeros(
                self.num_envs, dtype=torch.bool, device="cpu"
            )

        # Hidden States (assuming they are dictionaries of tensors)
        if self.recurrent and hidden_state is not None:
            # hidden_state is dict[str, Tensor] from PPO -> {key: (layers, num_envs, size)}
            # We need to populate current_step_data["hidden_states"]
            # with {key: (num_envs, layers, size)}
            current_step_data["hidden_states"] = {}
            for key, ppo_tensor_val in hidden_state.items():
                current_step_data["hidden_states"][key] = ppo_tensor_val.permute(
                    1, 0, 2
                )  # Shape: (num_envs, layers, size)

        # Create a TensorDict for the current step's data
        # This will have batch_size [num_envs]
        current_td_slice = TensorDict(
            current_step_data, batch_size=[self.num_envs], device="cpu"
        )

        # Assign this slice to the buffer at the current position
        # self.buffer has batch_size [capacity, num_envs]
        # self.buffer[self.pos] should be a TensorDict with batch_size [num_envs]
        # This assignment will correctly place the data, including the nested hidden_states TD.
        self.buffer[self.pos] = current_td_slice

        # Update buffer position
        self.pos += 1
        if self.pos == self.capacity:
            self.full = True

    def compute_returns_and_advantages(
        self, last_value: ArrayOrTensor, last_done: ArrayOrTensor
    ) -> None:
        """
        Compute returns and advantages for the stored experiences using GAE or Monte Carlo.

        :param last_value: Value estimate for the last observation in each environment (shape: (num_envs,))
        :type last_value: ArrayOrTensor
        :param last_done: Done flag for the last state in each environment (shape: (num_envs,))
        :type last_done: ArrayOrTensor
        """
        # Convert inputs to numpy arrays if they are tensors, and ensure correct shape
        if isinstance(last_value, torch.Tensor):
            last_value_np = last_value.cpu().numpy().reshape(self.num_envs)
        else:
            last_value_np = np.asarray(last_value).reshape(self.num_envs)

        if isinstance(last_done, torch.Tensor):
            last_done_np = last_done.cpu().numpy().reshape(self.num_envs)
        else:
            last_done_np = np.asarray(last_done).reshape(self.num_envs)

        buffer_size = self.capacity if self.full else self.pos

        # Temporary numpy arrays for computation, will be assigned back to TensorDict
        advantages_np = np.zeros((buffer_size, self.num_envs), dtype=np.float32)
        returns_np = np.zeros((buffer_size, self.num_envs), dtype=np.float32)

        # Get necessary data from TensorDict as numpy arrays for computation
        # Slicing to buffer_size for all components.
        rewards_np = self.buffer["rewards"][:buffer_size].cpu().numpy()
        dones_np = self.buffer["dones"][:buffer_size].cpu().numpy()
        values_np = self.buffer["values"][:buffer_size].cpu().numpy()

        if self.use_gae:
            last_gae_lambda = np.zeros(self.num_envs, dtype=np.float32)
            for t in reversed(range(buffer_size)):
                if t == buffer_size - 1:
                    next_non_terminal = 1.0 - last_done_np.astype(float)
                    next_values = last_value_np.astype(float)
                else:
                    next_non_terminal = 1.0 - dones_np[t + 1].astype(float)
                    next_values = values_np[t + 1]

                delta = (
                    rewards_np[t]
                    + self.gamma * next_values * next_non_terminal
                    - values_np[t]
                )
                advantages_np[t] = last_gae_lambda = (
                    delta
                    + self.gamma * self.gae_lambda * next_non_terminal * last_gae_lambda
                )
            returns_np = advantages_np + values_np
        else:
            # Monte Carlo returns
            last_returns_np = last_value_np.astype(float) * (
                1.0 - last_done_np.astype(float)
            )
            for t in reversed(range(buffer_size)):
                returns_np[t] = last_returns_np = rewards_np[
                    t
                ] + self.gamma * last_returns_np * (1.0 - dones_np[t].astype(float))
            advantages_np = returns_np - values_np

        # Assign computed advantages and returns back to the TensorDict
        self.buffer["advantages"][:buffer_size] = torch.from_numpy(advantages_np)
        self.buffer["returns"][:buffer_size] = torch.from_numpy(returns_np)

    def get(
        self, batch_size: Optional[int] = None
    ) -> dict[str, Union[np.ndarray, dict[str, np.ndarray]]]:
        """
        Get data from the buffer, flattened and optionally sampled into minibatches.

        :param batch_size: Size of the minibatch to sample. If None, returns all data. Defaults to None.
        :type batch_size: Optional[int]
        :return: Dictionary containing flattened buffer data arrays.
        :rtype: dict[str, Union[np.ndarray, dict[str, np.ndarray]]]
        """
        buffer_size = self.capacity if self.full else self.pos
        total_samples = buffer_size * self.num_envs

        if total_samples == 0:
            # Return empty dictionary or raise error if buffer is empty
            return {}

        # Get a view of the buffer up to the current position and for all envs
        # This slice will have batch_size [buffer_size, num_envs]
        valid_buffer_data: TensorDict = self.buffer[:buffer_size]

        # Reshape to flatten the num_envs dimension into the first batch dimension
        # New batch_size will be [buffer_size * num_envs]
        flattened_td = valid_buffer_data.view(-1)

        if batch_size is not None:
            if batch_size > total_samples:
                warnings.warn(
                    f"Batch size {batch_size} is larger than buffer size {total_samples}. Returning all data."
                )
                # Convert the flattened TensorDict to the old dictionary format
                # For hidden_states, we need to handle the nested TensorDict structure
                return self._convert_td_to_np_dict(flattened_td)

            indices = np.random.choice(total_samples, size=batch_size, replace=False)
            sampled_td = flattened_td[indices]
            # Convert the sampled TensorDict to the old dictionary format
            return self._convert_td_to_np_dict(sampled_td)
        else:
            return self._convert_td_to_np_dict(flattened_td)

    def get_tensor_batch(
        self, batch_size: Optional[int] = None, device: Optional[str] = None
    ) -> dict[str, Union[torch.Tensor, dict[str, torch.Tensor]]]:
        """
        Get data from the buffer as PyTorch tensors, flattened and optionally sampled.
        The output is a TensorDict.

        :param batch_size: Size of batch to sample, if None returns all data, defaults to None.
        :type batch_size: Optional[int]
        :param device: Device to put tensors on, defaults to None (uses self.device).
        :type device: Optional[str]
        :return: TensorDict containing the data.
        :rtype: dict[str, Union[torch.Tensor, dict[str, torch.Tensor]]]
        """
        target_device = device or self.device
        buffer_size = self.capacity if self.full else self.pos
        total_samples = buffer_size * self.num_envs

        if total_samples == 0:
            # Return an empty TensorDict with the correct device and structure if possible,
            # or simply an empty dict if structure is complex to define when empty.
            # For now, returning an empty dict to match previous behavior.
            return {}

        # Get a view of the buffer up to the current position and for all envs
        # This slice will have batch_size [buffer_size, num_envs]
        # All tensors inside self.buffer are already torch.Tensors on CPU
        valid_buffer_data_view: TensorDict = self.buffer[:buffer_size]

        # Reshape to flatten the num_envs dimension into the first batch dimension
        # New batch_size will be [buffer_size * num_envs]
        # .view(-1) is crucial for not creating a copy if possible
        flattened_td: TensorDict = valid_buffer_data_view.view(-1)

        if batch_size is not None:
            if batch_size > total_samples:
                warnings.warn(
                    f"Batch size {batch_size} is larger than buffer_size {total_samples}. "
                    "Returning all data."
                )
                # Move the whole flattened_td to the target device
                return flattened_td.to(target_device)

            indices = torch.randperm(total_samples, device="cpu")[
                :batch_size
            ]  # Sample on CPU then move
            sampled_td: TensorDict = flattened_td[indices]

            return sampled_td.to(target_device)
        else:
            # Return all flattened data, moved to the target device
            return flattened_td.to(target_device)

    # ------------------------------------------------------------------
    # New helper functions for truncated Backpropagation Through Time (BPTT)
    # ------------------------------------------------------------------

    def _convert_td_to_np_dict(
        self, td: TensorDict
    ) -> dict[str, Union[np.ndarray, dict[str, np.ndarray]]]:
        """
        Convert a TensorDict to a dictionary of numpy arrays.

        :param td: TensorDict to convert.
        :type td: TensorDict
        :return: Dictionary of numpy arrays.
        :rtype: dict[str, Union[np.ndarray, dict[str, np.ndarray]]]
        """
        # Convert the TensorDict to the old dictionary of numpy arrays format
        np_dict = {}
        for key, value in td.items():
            if key == "hidden_states" and isinstance(value, TensorDict):
                # For hidden states, we want the *initial* hidden state for each sequence.
                # The `value` here would be a TensorDict where each key (e.g., "h_actor")
                # has a tensor of shape (actual_batch_size, seq_len, layers, hidden_size)
                # We need to take the first time step [:, 0]
                # This block will be superseded by the explicit addition of "initial_hidden_states" below
                # but kept for structural reference or if full hidden sequences were needed in np_dict form.
                np_dict[key] = {
                    k_hid: v_hid.cpu().numpy() for k_hid, v_hid in value.items()
                }
            elif key in ["observations", "next_observations"] and isinstance(
                value, dict
            ):
                # Handle Dict observation spaces
                np_dict[key] = {
                    k_obs: v_obs.cpu().numpy() for k_obs, v_obs in value.items()
                }
            elif isinstance(
                value, TensorDict
            ):  # For nested like observations if it's a Dict space
                np_dict[key] = {
                    k_sub: v_sub.cpu().numpy() for k_sub, v_sub in value.items()
                }
            else:
                np_dict[key] = value.cpu().numpy()
        return np_dict

    @staticmethod
    def _pad_sequences(
        sequences: list[Union[torch.Tensor, TensorDict]],
        target_length: Optional[int] = None,
    ) -> Union[torch.Tensor, TensorDict]:
        """Pads sequences using zeros. If target_length is provided, the sequences are padded to the target length.
        Otherwise, the sequences are padded to the length of the longest sequence. Results in a tensor of
        shape (B, T, *). We provide the option to pad to a specified target length but in general these should be padded
        to the maximum length of the sequences in the batch (i.e. using `torch.nn.utils.rnn.pad_sequence`).

        :param sequences: The sequences to be padded.
        :type sequences: list[Union[torch.Tensor, TensorDict]]
        :param target_length: The target length to pad the sequences to. If None, the sequences are padded to the length of the longest sequence.
        :type target_length: Optional[int]
        :return: The padded sequence.
        :rtype: Union[torch.Tensor, TensorDict]
        """
        # Handle nested tensors
        if is_tensor_collection(sequences[0]):
            sequences_T = {
                key: [nested_seq[key] for nested_seq in sequences]
                for key in sequences[0].keys()
            }
            return TensorDict(
                {
                    key: RolloutBuffer._pad_sequences(sequences_T[key], target_length)
                    for key in sequences_T.keys()
                }
            )

        # If target_length is provided, pad the sequences to the target length
        if target_length is not None:
            padded_sequences = []
            for seq in sequences:
                current_length = seq.size(0)
                pad_amount = target_length - current_length
                pad_spec = [0, 0] * (seq.dim() - 1) + [0, pad_amount]
                padded_sequences.append(
                    torch.nn.functional.pad(seq, pad_spec, mode="constant", value=0)
                )

            return torch.stack(padded_sequences)

        # Otherwise, pad the sequences to the length of the longest sequence
        return torch.nn.utils.rnn.pad_sequence(
            sequences, batch_first=True, padding_value=0
        )

    def _get_complete_sequences(
        self,
        data: torch.Tensor,
        episode_done_indices: list[list[int]],
    ) -> tuple[list[torch.Tensor], int]:
        """Splits the provided data into sequences. If `self.max_seq_len` is not set, the entire episode
        is used as a sequence. Otherwise, the episode is split into sequences of length `self.max_seq_len`.
        If the episode is shorter than `self.max_seq_len`, the entire episode is used as a sequence.

        :param data: The data to be split into sequences.
        :type data: torch.Tensor
        :param episode_done_indices: The indices of done signals.
        :type episode_done_indices: list[list[int]]
        :return: A list of sequences and the length of the longest sequence.
        :rtype: tuple[list[torch.Tensor], int]
        """
        max_seq_len = self.max_seq_len
        sequences = []
        max_length = 1
        for env_idx in range(self.num_envs):
            start_index = 0
            for done_index in episode_done_indices[env_idx]:
                # Split trajectory into episodes (clone to make them independent for stack)
                episode = data[start_index : done_index + 1, env_idx]

                # Split episodes into sequences for truncated BPTT
                # If max_seq_len is not set, use the entire episode as a sequence
                # NOTE: It may be the case that we provide a max_seq_len but we have episodes
                # that are shorter than max_seq_len. In this case, we will use the entire episode
                # as a sequence, and later pad the shorter episodes to the max_seq_len.
                if (max_seq_len is not None) and (len(episode) >= max_seq_len):
                    # Extract sequences from the episode
                    truncated_sequences = extract_sequences_from_episode(
                        episode=episode,
                        max_seq_len=max_seq_len,
                        sequence_type=self.bptt_sequence_type,
                    )
                    sequences.extend(truncated_sequences)
                else:
                    sequences.append(episode)

                max_length = len(episode) if len(episode) > max_length else max_length
                start_index = done_index + 1

        return sequences, max_length

    def prepare_sequence_tensors(self, device: Optional[str] = None) -> TensorDict:
        """Returns a TensorDict with all of the possible sequences in the buffer for the
        observations, actions, and hidden states. We pad the sequences to the same length to obtain
        a TensorDict with batch_size [num_sequences, max_sequence_length] for efficient truncated BPTT.

        :param device: Device to put tensors on, defaults to None (uses self.device).
        :type device: Optional[str]
        :return: Dictionary with tensor sequences.
        :rtype: TensorDict
        """
        if not self.recurrent:
            raise ValueError(
                "prepare_sequence_tensors() is only supported when recurrent=True."
            )

        if self.size() == 0:
            raise ValueError("Attempting to prepare sequences with empty buffer.")

        target_device = device or self.device
        buffer_size = self.capacity if self.full else self.pos

        # Get a view of the buffer up to the current position and for all envs
        # This slice will have batch_size [buffer_size, num_envs]
        # All tensors inside self.buffer are already torch.Tensors on CPU
        valid_buffer_data_view: TensorDict = self.buffer[:buffer_size]

        # Split data into sequences and apply zero-padding
        # Retrieve the indices of dones as these are the last step of a whole episode
        episode_done_indices: list[list[int]] = []
        for env_idx in range(self.num_envs):
            env_dones: torch.Tensor = valid_buffer_data_view["dones"][:, env_idx]
            env_dones_list = env_dones.nonzero().squeeze(-1).tolist()
            episode_done_indices.append(env_dones_list)

            # Mark the end of the buffer as the end of an episode
            if (
                len(episode_done_indices[env_idx]) == 0
                or episode_done_indices[env_idx][-1] != buffer_size - 1
            ):
                episode_done_indices[env_idx].append(buffer_size - 1)

        # Get the indices of unpadded sequences
        flat_timesteps = torch.arange(buffer_size * self.num_envs).reshape(
            buffer_size, self.num_envs
        )
        unpadded_slices, _ = self._get_complete_sequences(
            flat_timesteps, episode_done_indices
        )

        self.unpadded_slices = unpadded_slices

        valid_data_to_pad: TensorDictBase = valid_buffer_data_view.select(
            "observations", "actions", "hidden_states"
        ).clone()

        valid_data_to_pad["pad_mask"] = torch.ones(
            valid_data_to_pad.batch_size, dtype=torch.bool
        )

        # Create a TensorDict with all of the possible sequences, padded to the same length
        padded_data_source = OrderedDict()
        for key, value in valid_data_to_pad.items():
            # Split data into episodes or sequences
            sequences, max_sequence_length = self._get_complete_sequences(
                value, episode_done_indices
            )

            # Apply zero-padding to ensure that each episode has the same length
            # NOTE: In https://github.com/MarcoMeter/recurrent-ppo-truncated-bpttm sequences
            # are padded to the maximum episode length, whereas these should be padded to the
            # specified max_sequence_length for truncated BPTT to see its benefits.
            padded_data_source[key] = RolloutBuffer._pad_sequences(
                sequences, target_length=None
            )

        self.max_sequence_length = (
            min(max_sequence_length, self.max_seq_len)
            if self.max_seq_len is not None
            else max_sequence_length
        )
        self.num_sequences = len(sequences)

        padded_td = TensorDict(
            source=padded_data_source,
            batch_size=[self.num_sequences, self.max_sequence_length],
        )

        # Now, create the "initial_hidden_states" part for the output TensorDict
        # This will be a nested TensorDict with batch_size [actual_batch_size]
        if self.recurrent and padded_td.get("hidden_states", None) is not None:
            initial_hidden_states = OrderedDict()
            full_hidden_sequences: TensorDict = padded_td.get("hidden_states")
            for h_key, h_val_tensor_sequences in full_hidden_sequences.items():
                initial_hidden_states[h_key] = h_val_tensor_sequences[:, 0].clone()

            # Use set_non_tensor for the dictionary of initial hidden state tensors
            padded_td.set_non_tensor("initial_hidden_states", initial_hidden_states)

            # Exclude original full hidden_states sequence
            padded_td = padded_td.exclude("hidden_states")

        # Select unpadded data for training
        unpadded_data_source: TensorDict = valid_buffer_data_view.select(
            "log_probs", "advantages", "values", "returns"
        )

        # Flatten the TensorDict's and move to the target device
        self.padded_data = padded_td.view(-1).to(target_device)
        self.unpadded_data = unpadded_data_source.view(-1).to(target_device)

    def get_minibatch_sequences(
        self,
        batch_size: int,
    ) -> Generator[tuple[StandardTensorDict, StandardTensorDict], None, None]:
        """Get a minibatch of sequences from the buffer.

        :param batch_size: The number of sequences to sample.
        :type batch_size: int
        :return: A TensorDict containing the minibatch of sequences.
        :rtype: Generator[tuple[dict[str, torch.Tensor], dict[str, torch.Tensor]], None, None]
        """
        if self.unpadded_slices is None:
            raise ValueError(
                "Attempting to fetch minibatches before preparing sequences. "
                "Call prepare_sequence_tensors() first."
            )

        # Determine the number of sequences per mini batch
        if batch_size > self.num_sequences:
            warnings.warn(
                f"Batch size {batch_size} is larger than the number of sequences "
                f"({self.num_sequences}), using batch_size = {self.num_sequences}."
            )
            batch_size = self.num_sequences

        num_batches = self.num_sequences // batch_size
        num_sequences_per_batch = [batch_size] * num_batches
        remainder = self.num_sequences % batch_size

        if remainder > 0:
            num_sequences_per_batch.append(remainder)

        # Prepare indices, but only shuffle the sequence indices and not the
        # entire batch to ensure that sequences are maintained as a whole.
        indices = torch.arange(
            start=0,
            end=self.num_sequences * self.max_sequence_length,
        ).reshape(self.num_sequences, self.max_sequence_length)

        sequence_indices = torch.randperm(self.num_sequences)

        # Compose mini batches
        start = 0
        for num_sequences in num_sequences_per_batch:
            end = start + num_sequences
            sequences_samples = sequence_indices[start:end]
            padded_indices = indices[sequences_samples].view(-1)

            # Unpadded and flat indices are used to sample unpadded training data
            minibatch_seq_indices: list[int] = sequences_samples.tolist()
            unpadded_indices = [
                self.unpadded_slices[idx].tolist() for idx in minibatch_seq_indices
            ]
            unpadded_indices = [
                item for sublist in unpadded_indices for item in sublist
            ]

            padded: TensorDict = self.padded_data[padded_indices]
            if self.recurrent and padded.get("initial_hidden_states", None) is not None:
                batch_hidden_states = {}
                initial_hidden_states: TensorDict = padded.get_non_tensor(
                    "initial_hidden_states"
                )
                for key, value in initial_hidden_states.items():
                    batch_hidden_states[key] = value[sequences_samples].clone()

                padded.set_non_tensor("initial_hidden_states", batch_hidden_states)

            unpadded = self.unpadded_data[unpadded_indices]
            start = end
            yield padded, unpadded

    def __getstate__(self) -> dict[str, Any]:
        """Gets the state dictionary for pickling, ensuring arrays are copied.

        :return: State dictionary.
        :rtype: dict[str, Any]
        """
        return self.__dict__.copy()

    def __setstate__(self, state: dict[str, Any]) -> None:
        """Sets the state dictionary when unpickling, re-initializing buffers.

        :param state: State dictionary.
        :type state: dict[str, Any]
        """
        self.__dict__.update(state)

        # Let's assume self.buffer is correctly loaded by `self.__dict__.update(state)`.
        # We should verify its integrity or re-initialize if it's somehow corrupted/not present.
        if not hasattr(self, "buffer") or not isinstance(self.buffer, TensorDict):
            warnings.warn(
                "TensorDict buffer not found or invalid during unpickling. Re-initializing."
            )
            if not hasattr(self, "observation_space") or not hasattr(
                self, "action_space"
            ):
                warnings.warn(
                    "Observation or action space missing during RolloutBuffer unpickling. Buffer might be invalid."
                )
                return

            self._initialize_buffers()
            self.buffer = self.buffer.to("cpu")  # Ensure the buffer is on CPU
        else:
            # Verify batch_size and keys if necessary
            expected_batch_size = torch.Size([self.capacity, self.num_envs])
            if self.buffer.batch_size != expected_batch_size:
                warnings.warn(
                    f"Loaded TensorDict has batch_size {self.buffer.batch_size}, expected {expected_batch_size}. Re-initializing."
                )
                self._initialize_buffers()

            # Ensure the loaded buffer is on the correct device (CPU) as expected by internal logic
            self.buffer = self.buffer.to("cpu")
